/**
 * com.shorindo.FtpConnection - an implementation of FTP client.
 *      $Revision: 1.7 $
 *      $Date: 2001/07/17 01:57:27 $
 *      $Author: kazm $
 * Copyright (C) 2001 Shorindo, Inc.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 */

package com.shorindo;
import java.io.*;
import java.net.*;
import java.util.*;

/**
 * ñFTP饤ȤμŵŪˤϰʲΤ褦˻Ѥޤ
 * <PRE>
 *	FtpConnection ftp = new FtpConnection();
 *
 *	ftp.login("user", "pass");
 *	ftp.cd("/tmp");
 *	ftp.put("a.txt", "b.txt");
 *	ftp.get("b.txt", "c.txt");
 *	ftp.quit();
 * </PRE>
 * @author kazm@shorindo.com
 * @version $Revision: 1.7 $
 */
public class FtpConnection {
    private static final int CTRLPORT = 21;
    private static final int TIMEOUT = 10000;    // msec
    private static final int STAT_INIT    = 0;
    private static final int STAT_CONN    = 1;
    private static final int STAT_COMMAND = 2;
    private static final int STAT_WAITRESP= 3;
    private static final int TYPE_ASCII   = 0;
    private static final int TYPE_EBCDIC  = 1;
    private static final int TYPE_IMAGE   = 2;

    private int debug = 0;
    private int status = STAT_INIT;
    private int type = TYPE_ASCII;
    private int timer = TIMEOUT;
    private Socket ctrlSock;
    private BufferedReader ctrlIn;
    private PrintWriter ctrlOut;
    private Socket dataSock;
    private InputStream dataIn;
    private OutputStream dataOut;
    private byte inetaddr[] = null;

    /**
     * FTPΥå饹Υ󥹥ȥ饯
     */
    public FtpConnection () {
    }

    /**
     * ꤷFTPФȤδ֤FTPå򳫻Ϥޤ
     * @param host FTPФΥۥ̾⤷IPɥ쥹
     */
    public void open (String host) throws IOException {
        String resp;

        ctrlSock  = new Socket(host, CTRLPORT);
        ctrlSock.setSoTimeout(timer);
        ctrlIn    = new BufferedReader(
                new InputStreamReader(
                        ctrlSock.getInputStream()));
        ctrlOut   = new PrintWriter(ctrlSock.getOutputStream());
        resp = waitResponse();
    }

    /**
     * FTPФФFTPޥɤȯԤ쥹ݥ󥹤ޤ
     * @param command FTPФФꥯȥޥɡʥѥ᡼ޤ
     * @return FTPФΥ쥹ݥʸʥɤޤ
     */
    private String doCommand (String command) throws IOException {
        String resp = null;

        ctrlOut.print(command + "\r\n");
        ctrlOut.flush();
        if (debug > 0) System.err.println("> " + command);
        status = STAT_COMMAND;
        resp = waitResponse();

        return resp;
    }

    /**
     * FTPФΥ쥹ݥ󥹤ޤ
     * 쥹ݥ󥹥ɤ400ְʾξIOExceptionꤲޤ
     * ΤԤ˥ॢȤȯϡ
     * SocketExceptionꤲޤ
     * @return FTPФΥ쥹ݥʸʥɤޤ
     */
    private String waitResponse () throws IOException {
        String resp = "";
        String line;
        int code;

        while (true) {
            line = ctrlIn.readLine();
            resp += line;
	    if (debug > 0) System.err.println("< " + line);
            if (line.length() >= 4 &&
                line.substring(3, 4).equals(" "))
                break;
        }

        code = Integer.parseInt(line.substring(0, 3));
        if (code >= 400)
            throw new IOException(resp);

        return resp;
    }

    /**
     * Ω줿FTPåФơǧڤԤޤ
     * @param user 桼ID
     * @param pass ѥ
     */
    public void login (String user, String pass) throws IOException {
        doCommand("USER " + user);
        doCommand("PASS " + pass);
    }

    /**
     * ߤFTPåλޤ
     */
    public void quit () throws IOException {
        doCommand("QUIT");
        if (ctrlSock != null) ctrlSock.close();
        if (dataSock != null) dataSock.close();
    }

    /**
     * FTPФNLSTޥɤȯԤߤΥ⡼ȥǥ쥯ȥˤ
     * ե̾ΰ֤ޤ
     * @return ե̾ޤ
     */
    public String[] ls () throws IOException {
        return ls(null);
    }

    /**
     * FTPФNLSTޥɤȯԤꤷǥ쥯ȥ⤷ϥե̾
     * ΰ֤ޤ*פʤɤΥ磻ɥɤŸϹԤޤΤǡ
     * ե̾ꤷƤ⡢Ʊ̾Υե֤̾ʤΤǰ̣
     * ޤ
     * @param dir ǥ쥯ȥ̾ޤϥե̾
     * @return ե̾ޤ
     */
    public String[] ls (String dir) throws IOException {
        String resp[] = null;
        String cmd = "";
	Vector v = new Vector();
        byte buff[] = new byte[1024];
        int len = 0;
        int currentType = type;

        ascii();

        if (inetaddr == null)
            inetaddr = InetAddress.getLocalHost().getAddress();
        ServerSocket servSock = new ServerSocket(0, 1);
        for (int i = 0; i < 4; i++)
            cmd += (inetaddr[i]&0xff) + ",";

        doCommand("PORT " + cmd + ((servSock.getLocalPort()/256)&0xff) +
                  "," + ((servSock.getLocalPort()%256)&0xff));

        if (dir == null)
            doCommand("NLST");
        else
            doCommand("NLST " + dir);
        dataSock = servSock.accept();
        dataSock.setSoTimeout(timer);
        servSock.close();
        dataIn = dataSock.getInputStream();
	BufferedReader dataReader = new BufferedReader(
		new InputStreamReader(dataIn));
	String file;
        while ((file = dataReader.readLine()) != null)
	    v.addElement(file);
        dataIn.close();
        waitResponse();

        if (currentType == TYPE_ASCII) ascii();
        else if (currentType == TYPE_IMAGE) binary();
        else if (currentType == TYPE_EBCDIC);   // not support

	resp = new String[v.size()];
	for (int i = 0; i < resp.length; i++)
	  resp[i] = (String)v.elementAt(i);
        return resp;
    }

    /**
     * FTPоθߤΥǥ쥯ȥѹޤ
     * ľϥ桼Υۡǥ쥯ȥǤFTPФˤ
     * ۤʤޤ
     * @param dir ưΥǥ쥯ȥ̾
     */
    public void cd (String dir) throws IOException {
        doCommand("CWD " + dir);
    }

    /**
     * FTPо˿ǥ쥯ȥޤ
     * ꤵ줿̾¸ߤƤʤɤͳˤꡢǥ쥯ȥ꤬
     * ǤʤäϡIOExceptionꤲޤ
     * ºݤˤɤΤ褦ʤȤˤ㳰ꤲ뤫ϡFTPФ֤쥹ݥ
     * ˰¸ޤ
     * @param dir ǥ쥯ȥ̾
     */
    public void mkdir (String dir) throws IOException {
        doCommand("MKD " + dir);
    }

    /**
     * FTPоΥǥ쥯ȥޤ
     * ̾Ϻ˼ԤȤϡIOExceptionꤲޤFTPФ֤
     * 쥹ݥ󥹤˰¸ޤ
     * @param dir ǥ쥯ȥ̾
     */
    public void rmdir (String dir) throws IOException {
        doCommand("RMD " + dir);
    }

    /**
     * FTPоθߤΥǥ쥯ȥ֤ޤ
     * ֤ͤϥ쥹ݥ󥹥ɤޤߡϥФ˰¸ޤ
     * @return ߤΥǥ쥯ȥ
     */
    public String pwd () throws IOException {
        String resp = doCommand("PWD");
        return resp;
    }

    /**
     * FTPФ˥եޤ
     * @param rfile FTPоΥե̾
     * @param lfile ե̾
     */
    public void get (String rfile, String lfile) throws IOException {
        String cmd = "";
        FileOutputStream fos;
        byte buff[] = new byte[1024];
        int len = 0;

        if (inetaddr == null)
            inetaddr = InetAddress.getLocalHost().getAddress();
        ServerSocket servSock = new ServerSocket(0, 1);
        for (int i = 0; i < 4; i++)
            cmd += (inetaddr[i]&0xff) + ",";

        doCommand("PORT " + cmd + ((servSock.getLocalPort()/256)&0xff) +
                  "," + ((servSock.getLocalPort()%256)&0xff));

        doCommand("RETR " + rfile);
        dataSock = servSock.accept();
        servSock.close();
        dataIn = dataSock.getInputStream();
        fos = new FileOutputStream(lfile);
        while ((len = dataIn.read(buff)) > 0) {
            fos.write(buff, 0, len);
        }
        fos.close();
        dataIn.close();
        waitResponse();
    }

    /**
     * 뤫FTPФ˥եޤ
     * @param lfile ե̾
     * @param rfile FTPоΥե̾
     */
    public void put (String lfile, String rfile) throws IOException {
        String cmd = "";
        FileInputStream fis;
        byte buff[] = new byte[1024];
        int len = 0;

        if (inetaddr == null)
            inetaddr = InetAddress.getLocalHost().getAddress();
        ServerSocket servSock = new ServerSocket(0, 1);
        for (int i = 0; i < 4; i++)
            cmd += (inetaddr[i]&0xff) + ",";

        doCommand("PORT " + cmd + ((servSock.getLocalPort()/256)&0xff) +
                  "," + ((servSock.getLocalPort()%256)&0xff));

        doCommand("STOR " + rfile);
        dataSock = servSock.accept();
        servSock.close();
        dataOut = dataSock.getOutputStream();
        fis = new FileInputStream(lfile);
        while ((len = fis.read(buff)) > 0) {
            dataOut.write(buff, 0, len);
        }
        fis.close();
        dataOut.close();
        waitResponse();
    }

    /**
     * ߤǤޤ
     * εǽ̤бǡºݤˤϲⵯޤ
     */
    public void abort () throws IOException {
    }

    /**
     * ž⡼ɤASCIIˤޤ
     */
    public void ascii () throws IOException {
        doCommand("TYPE A");
        type = TYPE_ASCII;
    }

    /**
     * ž⡼ɤBINARYˤޤ
     */
    public void binary () throws IOException {
        doCommand("TYPE I");
        type = TYPE_IMAGE;
    }

    /**
     * θߤΥǥ쥯ȥưޤ
     * εǽ̤бǡºݤˤϲⵯޤ
     */
    public void lcd (String dir) throws IOException {
    }

    /**
     * FTPоΥեޤ
     * @param file FTPоΥե̾
     */
    public void delete (String file) throws IOException {
        doCommand("DELE " + file);
    }

    /**
     * FTPоΥե̾ѹޤ
     * @param oldfile FTPоθߤΥե̾
     * @param newfile FTPоοե̾
     */
    public void rename (String oldfile, String newfile) throws IOException {
        doCommand("RNFR " + oldfile);
        doCommand("RNTO " + newfile);
    }

    /**
     * ʣΥͥåȥ󥿥եĴĶǡѤ륤󥿥ե
     * IPɥ쥹ꤷޤ
     * @param h1 IPɥ쥹ΣХ
     * @param h2 IPɥ쥹ΣХ
     * @param h3 IPɥ쥹ΣХ
     * @param h4 IPɥ쥹ΣХ
     */
    public void setInterface (int h1, int h2, int h3, int h4) {
        inetaddr = new byte[4];
        inetaddr[0] = (byte)(h1&0xff);
        inetaddr[1] = (byte)(h2&0xff);
        inetaddr[2] = (byte)(h3&0xff);
        inetaddr[3] = (byte)(h4&0xff);
    }

    /**
     * åȤɽФԤ֤ꤷޤ
     * Ԥ֤Υǥեͤ1000ߥäǤ
     * @param timer Ԥ(ߥ)
     */
    public void setTimeout (int timer) {
	this.timer = timer;
    }

    /**
     * ǥХå⡼ɤꤷޤ
     * ǥХå٥0礭ͤꤹȡץȥΥȥ졼
     * ɸϤ˽Ϥޤ
     * @param level ǥХå٥
     */
    public void debug (int level) {
        debug = level;
    }

    public void finalize () throws Throwable {
        quit();
        super.finalize();
    }
}

